@@ -1,18 +1,21 @@
-{-# LANGUAGE ViewPatterns #-}
+{-# LANGUAGE ImplicitParams, ViewPatterns, NoMonomorphismRestriction #-}
 import Control.Concurrent
 
 import Control.Applicative ((<$))
 import Control.Concurrent (threadDelay)
 import Control.Exception (try)
+import qualified Data.ByteString.Lazy.Char8 as BL (pack)
 import Data.Char (isSpace)
+import qualified Data.Digest.Pure.MD5 as MD5 (md5)
 import Data.List.Split (splitOn)
 import Data.Maybe (fromJust)
 import System.IO.Error (userError)
 
-import qualified Network.Shed.Httpd as Httpd
+import qualified Httpd
 
 import Network.Browser
 import Network.HTTP
+import Network.HTTP.Base
 import Network.HTTP.Auth
 import Network.HTTP.Headers
 import Network.Stream (Result)
@@ -26,55 +29,73 @@ import Test.Framework.Providers.HUnit
 import Test.HUnit
 
 
-basicGetRequest :: Assertion
+basicGetRequest :: (?testUrl :: ServerAddress) => Assertion
 basicGetRequest = do
-  response <- simpleHTTP (getRequest (testUrl "/basic/get"))
+  response <- simpleHTTP (getRequest (?testUrl "/basic/get"))
   code <- getResponseCode response
   assertEqual "HTTP status code" (2, 0, 0) code
   body <- getResponseBody response
   assertEqual "Receiving expected response" "It works." body
 
-basicExample :: Assertion
+basicGetRequestLBS :: (?testUrl :: ServerAddress) => Assertion
+basicGetRequestLBS = do
+  response <- simpleHTTP (mkRequest GET (fromJust (parseURI (?testUrl ("/basic/get")))))
+  code <- getResponseCode response
+  assertEqual "HTTP status code" (2, 0, 0) code
+  body <- getResponseBody response
+  assertEqual "Receiving expected response" (BL.pack "It works.") body
+
+basicHeadRequest :: (?testUrl :: ServerAddress) => Assertion
+basicHeadRequest = do
+  response <- simpleHTTP (headRequest (?testUrl "/basic/head"))
+  code <- getResponseCode response
+  assertEqual "HTTP status code" (2, 0, 0) code
+  body <- getResponseBody response
+  -- the body should be empty, since this is a HEAD request
+  assertEqual "Receiving expected response" "" body
+
+basicExample :: (?testUrl :: ServerAddress) => Assertion
 basicExample = do
   result <-
     -- sample code from Network.HTTP haddock, with URL changed
     -- Note there's also a copy of the example in the .cabal file
-    simpleHTTP (getRequest (testUrl "/basic/example")) >>= fmap (take 100) . getResponseBody
+    simpleHTTP (getRequest (?testUrl "/basic/example")) >>= fmap (take 100) . getResponseBody
   assertEqual "Receiving expected response" (take 100 haskellOrgText) result
 
-secureGetRequest :: Assertion
+secureGetRequest :: (?secureTestUrl :: ServerAddress) => Assertion
 secureGetRequest = do
-  response <- try $ simpleHTTP (getRequest (secureTestUrl "/anything"))
+  response <- try $ simpleHTTP (getRequest (?secureTestUrl "/anything"))
   assertEqual "Threw expected exception"
               (Left (userError "https not supported"))
               (fmap show response) -- fmap show because Response isn't in Eq
 
-basicPostRequest :: Assertion
+basicPostRequest :: (?testUrl :: ServerAddress) => Assertion
 basicPostRequest = do
   let sendBody = "body"
-  response <- simpleHTTP $ postRequestWithBody (testUrl "/basic/post")
+  response <- simpleHTTP $ postRequestWithBody (?testUrl "/basic/post")
                                                "text/plain"
                                                sendBody
   code <- getResponseCode response
   assertEqual "HTTP status code" (2, 0, 0) code
   body <- getResponseBody response
   assertEqual "Receiving expected response"
-              (show (Just "text/plain\r", Just "4\r", sendBody))
+              (show (Just "text/plain", Just "4", sendBody))
               body
 
-basicAuthFailure :: Assertion
+basicAuthFailure :: (?testUrl :: ServerAddress) => Assertion
 basicAuthFailure = do
-  response <- simpleHTTP (getRequest (testUrl "/auth/basic"))
+  response <- simpleHTTP (getRequest (?testUrl "/auth/basic"))
   code <- getResponseCode response
   body <- getResponseBody response
   assertEqual "HTTP status code" ((4, 0, 1), "Nothing") (code, body)
 
+credentialsBasic :: (?testUrl :: ServerAddress) => Authority
 credentialsBasic = AuthBasic "Testing realm" "test" "password"
-                             (fromJust . parseURI . testUrl $ "/auth/basic")
+                             (fromJust . parseURI . ?testUrl $ "/auth/basic")
 
-basicAuthSuccess :: Assertion
+basicAuthSuccess :: (?testUrl :: ServerAddress) => Assertion
 basicAuthSuccess = do
-  let req = getRequest (testUrl "/auth/basic")
+  let req = getRequest (?testUrl "/auth/basic")
   let authString = withAuthority credentialsBasic req
   let reqWithAuth = req { rqHeaders = mkHeader HdrAuthorization authString:rqHeaders req }
   response <- simpleHTTP reqWithAuth
@@ -82,7 +103,7 @@ basicAuthSuccess = do
   body <- getResponseBody response
   assertEqual "Receiving expected response" ((2, 0, 0), "Here's the secret") (code, body)
 
-browserExample :: Assertion
+browserExample :: (?testUrl :: ServerAddress) => Assertion
 browserExample = do
   result <-
     -- sample code from Network.Browser haddock, with URL changed
@@ -91,16 +112,16 @@ browserExample = do
       (_, rsp)
          <- Network.Browser.browse $ do
                setAllowRedirects True -- handle HTTP redirects
-               request $ getRequest (testUrl "/browser/example")
+               request $ getRequest (?testUrl "/browser/example")
       return (take 100 (rspBody rsp))
   assertEqual "Receiving expected response" (take 100 haskellOrgText) result
 
 -- A vanilla HTTP request using Browser shouln't send a cookie header
-browserNoCookie :: Assertion
+browserNoCookie :: (?testUrl :: ServerAddress) => Assertion
 browserNoCookie = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
-    request $ getRequest (testUrl "/browser/no-cookie")
+    request $ getRequest (?testUrl "/browser/no-cookie")
   let code = rspCode response
   assertEqual "HTTP status code" (2, 0, 0) code
 
@@ -113,53 +134,53 @@ browserNoCookie = do
 -- Expected: Server gets single cookie with "hello=world"
 -- Actual:   Server gets 3 extra cookies, which are actually cookie attributes:
 --           "$Version=0;hello=world;$Domain=localhost:8080\r"
-browserOneCookie :: Assertion
+browserOneCookie :: (?testUrl :: ServerAddress) => Assertion
 browserOneCookie = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
     setMaxPoolSize (Just 0) -- TODO remove this: workaround for github issue 14
     -- This first requests returns a single Set-Cookie: hello=world
-    _ <- request $ getRequest (testUrl "/browser/one-cookie/1")
+    _ <- request $ getRequest (?testUrl "/browser/one-cookie/1")
 
     -- This second request should send a single Cookie: hello=world
-    request $ getRequest (testUrl "/browser/one-cookie/2")
+    request $ getRequest (?testUrl "/browser/one-cookie/2")
   let body = rspBody response
   assertEqual "Receiving expected response" "" body
   let code = rspCode response
   assertEqual "HTTP status code" (2, 0, 0) code
 
-browserTwoCookies :: Assertion
+browserTwoCookies :: (?testUrl :: ServerAddress) => Assertion
 browserTwoCookies = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
     setMaxPoolSize (Just 0) -- TODO remove this: workaround for github issue 14
     -- This first request returns two cookies
-    _ <- request $ getRequest (testUrl "/browser/two-cookies/1")
+    _ <- request $ getRequest (?testUrl "/browser/two-cookies/1")
 
     -- This second request should send them back
-    request $ getRequest (testUrl "/browser/two-cookies/2")
+    request $ getRequest (?testUrl "/browser/two-cookies/2")
   let body = rspBody response
   assertEqual "Receiving expected response" "" body
   let code = rspCode response
   assertEqual "HTTP status code" (2, 0, 0) code
 
 
-browserFollowsRedirect :: Int -> Assertion
+browserFollowsRedirect :: (?testUrl :: ServerAddress) => Int -> Assertion
 browserFollowsRedirect n = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
     setMaxPoolSize (Just 0) -- TODO remove this: workaround for github issue 14
-    request $ getRequest (testUrl "/browser/redirect/relative/" ++ show n ++ "/basic/get")
+    request $ getRequest (?testUrl "/browser/redirect/relative/" ++ show n ++ "/basic/get")
   assertEqual "Receiving expected response from server"
               ((2, 0, 0), "It works.")
               (rspCode response, rspBody response)
 
-browserReturnsRedirect :: Int -> Assertion
+browserReturnsRedirect :: (?testUrl :: ServerAddress) => Int -> Assertion
 browserReturnsRedirect n = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
     setMaxPoolSize (Just 0) -- TODO remove this: workaround for github issue 14
-    request $ getRequest (testUrl "/browser/redirect/relative/" ++ show n ++ "/basic/get")
+    request $ getRequest (?testUrl "/browser/redirect/relative/" ++ show n ++ "/basic/get")
   assertEqual "Receiving expected response from server"
               ((n `div` 100, n `mod` 100 `div` 10, n `mod` 10), "")
               (rspCode response, rspBody response)
@@ -167,7 +188,7 @@ browserReturnsRedirect n = do
 authGenBasic _ "Testing realm" = return $ Just ("test", "password")
 authGenBasic _ realm = fail $ "Unexpected realm " ++ realm
 
-browserBasicAuth :: Assertion
+browserBasicAuth :: (?testUrl :: ServerAddress) => Assertion
 browserBasicAuth = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
@@ -175,7 +196,7 @@ browserBasicAuth = do
 
     setAuthorityGen authGenBasic
 
-    request $ getRequest (testUrl "/auth/basic")
+    request $ getRequest (?testUrl "/auth/basic")
 
   assertEqual "Receiving expected response from server"
               ((2, 0, 0), "Here's the secret")
@@ -184,7 +205,7 @@ browserBasicAuth = do
 authGenDigest _ "Digest testing realm" = return $ Just ("test", "digestpassword")
 authGenDigest _ realm = fail $ "Unexpected digest realm " ++ realm
 
-browserDigestAuth :: Assertion
+browserDigestAuth :: (?testUrl :: ServerAddress) => Assertion
 browserDigestAuth = do
   (_, response) <- browse $ do
     setOutHandler (const $ return ())
@@ -192,7 +213,7 @@ browserDigestAuth = do
 
     setAuthorityGen authGenDigest
 
-    request $ getRequest (testUrl "/auth/digest")
+    request $ getRequest (?testUrl "/auth/digest")
 
   assertEqual "Receiving expected response from server"
               ((2, 0, 0), "Here's the digest secret")
@@ -200,13 +221,13 @@ browserDigestAuth = do
 
 
 
-browserAlt :: Assertion
+browserAlt :: (?altTestUrl :: ServerAddress) => Assertion
 browserAlt = do
   (response) <- browse $ do
 
     setOutHandler (const $ return ())
 
-    (_, response1) <- request $ getRequest (altTestUrl "/basic/get")
+    (_, response1) <- request $ getRequest (?altTestUrl "/basic/get")
 
     return response1
 
@@ -216,13 +237,13 @@ browserAlt = do
 
 -- test that requests to multiple servers on the same host
 -- don't get confused with each other
-browserBoth :: Assertion
+browserBoth :: (?testUrl :: ServerAddress, ?altTestUrl :: ServerAddress) => Assertion
 browserBoth = do
   (response1, response2) <- browse $ do
     setOutHandler (const $ return ())
 
-    (_, response1) <- request $ getRequest (testUrl "/basic/get")
-    (_, response2) <- request $ getRequest (altTestUrl "/basic/get")
+    (_, response1) <- request $ getRequest (?testUrl "/basic/get")
+    (_, response2) <- request $ getRequest (?altTestUrl "/basic/get")
 
     return (response1, response2)
 
@@ -236,13 +257,13 @@ browserBoth = do
 
 -- test that requests to multiple servers on the same host
 -- don't get confused with each other
-browserBothReversed :: Assertion
+browserBothReversed :: (?testUrl :: ServerAddress, ?altTestUrl :: ServerAddress) => Assertion
 browserBothReversed = do
   (response1, response2) <- browse $ do
     setOutHandler (const $ return ())
 
-    (_, response2) <- request $ getRequest (altTestUrl "/basic/get")
-    (_, response1) <- request $ getRequest (testUrl "/basic/get")
+    (_, response2) <- request $ getRequest (?altTestUrl "/basic/get")
+    (_, response1) <- request $ getRequest (?testUrl "/basic/get")
 
     return (response1, response2)
 
@@ -254,49 +275,49 @@ browserBothReversed = do
               ((2, 0, 0), "This is the alternate server.")
               (rspCode response2, rspBody response2)
 
-browserSecureRequest :: Assertion
+browserSecureRequest :: (?secureTestUrl :: ServerAddress) => Assertion
 browserSecureRequest = do
   res <- try $ browse $ do
     setOutHandler (const $ return ())
 
-    request $ getRequest (secureTestUrl "/anything")
+    request $ getRequest (?secureTestUrl "/anything")
 
   assertEqual "Threw expected exception"
               (Left (userError "https not supported"))
               (fmap show res) -- fmap show because Response isn't in Eq
 
 -- in case it tries to reuse the connection
-browserSecureRequestAfterInsecure :: Assertion
+browserSecureRequestAfterInsecure :: (?testUrl :: ServerAddress, ?secureTestUrl :: ServerAddress) => Assertion
 browserSecureRequestAfterInsecure = do
   res <- try $ browse $ do
     setOutHandler (const $ return ())
 
-    request $ getRequest (testUrl "/basic/get")
-    request $ getRequest (secureTestUrl "/anything")
+    request $ getRequest (?testUrl "/basic/get")
+    request $ getRequest (?secureTestUrl "/anything")
 
   assertEqual "Threw expected exception"
               (Left (userError "https not supported"))
               (fmap show res) -- fmap show because Response isn't in Eq
 
-browserRedirectToSecure :: Assertion
+browserRedirectToSecure :: (?testUrl :: ServerAddress, ?secureTestUrl :: ServerAddress) => Assertion
 browserRedirectToSecure = do
   res <- try $ browse $ do
     setOutHandler (const $ return ())
     setErrHandler fail
 
-    request $ getRequest (testUrl "/browser/redirect/secure/301/anything")
+    request $ getRequest (?testUrl "/browser/redirect/secure/301/anything")
 
   assertEqual "Threw expected exception"
-              (Left (userError $ "Unable to handle redirect, unsupported scheme: " ++ secureTestUrl "/anything"))
+              (Left (userError $ "Unable to handle redirect, unsupported scheme: " ++ ?secureTestUrl "/anything"))
               (fmap show res) -- fmap show because Response isn't in Eq
 
-browserTwoRequests :: Assertion
+browserTwoRequests :: (?testUrl :: ServerAddress) => Assertion
 browserTwoRequests = do
   (response1, response2) <- browse $ do
     setOutHandler (const $ return ())
 
-    (_, response1) <- request $ getRequest (testUrl "/basic/get")
-    (_, response2) <- request $ getRequest (testUrl "/basic/get2")
+    (_, response1) <- request $ getRequest (?testUrl "/basic/get")
+    (_, response2) <- request $ getRequest (?testUrl "/basic/get2")
 
     return (response1, response2)
 
@@ -309,14 +330,14 @@ browserTwoRequests = do
               (rspCode response2, rspBody response2)
 
 
-browserTwoRequestsAlt :: Assertion
+browserTwoRequestsAlt :: (?altTestUrl :: ServerAddress) => Assertion
 browserTwoRequestsAlt = do
   (response1, response2) <- browse $ do
 
     setOutHandler (const $ return ())
 
-    (_, response1) <- request $ getRequest (altTestUrl "/basic/get")
-    (_, response2) <- request $ getRequest (altTestUrl "/basic/get2")
+    (_, response1) <- request $ getRequest (?altTestUrl "/basic/get")
+    (_, response2) <- request $ getRequest (?altTestUrl "/basic/get2")
 
     return (response1, response2)
 
@@ -328,15 +349,15 @@ browserTwoRequestsAlt = do
               ((2, 0, 0), "This is the alternate server (2).")
               (rspCode response2, rspBody response2)
 
-browserTwoRequestsBoth :: Assertion
+browserTwoRequestsBoth :: (?testUrl :: ServerAddress, ?altTestUrl :: ServerAddress) => Assertion
 browserTwoRequestsBoth = do
   (response1, response2, response3, response4) <- browse $ do
     setOutHandler (const $ return ())
 
-    (_, response1) <- request $ getRequest (testUrl "/basic/get")
-    (_, response2) <- request $ getRequest (altTestUrl "/basic/get")
-    (_, response3) <- request $ getRequest (testUrl "/basic/get2")
-    (_, response4) <- request $ getRequest (altTestUrl "/basic/get2")
+    (_, response1) <- request $ getRequest (?testUrl "/basic/get")
+    (_, response2) <- request $ getRequest (?altTestUrl "/basic/get")
+    (_, response3) <- request $ getRequest (?testUrl "/basic/get2")
+    (_, response4) <- request $ getRequest (?altTestUrl "/basic/get2")
 
     return (response1, response2, response3, response4)
 
@@ -384,100 +405,118 @@ haskellOrgText =
 \\t\t<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\
 \\t\t\t\t<meta name=\"keywords\" content=\"Haskell,Applications and libraries,Books,Foreign Function Interface,Functional programming,Hac Boston,HakkuTaikai,HaskellImplementorsWorkshop/2011,Haskell Communities and Activities Report,Haskell in education,Haskell in industry\" />"
 
-processRequest :: Httpd.Request -> IO Httpd.Response
+digestMatch
+  username realm password
+  nonce opaque
+  method relativeURI makeAbsolute
+  headers
+  =
+  common `isSubsetOf` headers && (relative `isSubsetOf` headers || absolute `isSubsetOf` headers)
+ where
+   common = [("username", show username), ("realm", show realm), ("nonce", show nonce),
+             ("opaque", show opaque)]
+   md5 = show . MD5.md5 . BL.pack
+   ha1 = md5 (username++":"++realm++":"++password)
+   ha2 uri = md5 (method++":"++uri)
+   response uri = md5 (ha1 ++ ":" ++ nonce ++ ":" ++ ha2 uri)
+   mkUncommon uri hash = [("uri", show uri), ("response", show hash)]
+   relative = mkUncommon relativeURI (response relativeURI)
+   absoluteURI = makeAbsolute relativeURI
+   absolute = mkUncommon absoluteURI (response absoluteURI)
+
+processRequest :: (?testUrl :: ServerAddress, ?secureTestUrl :: ServerAddress)
+               => Httpd.Request
+               -> IO Httpd.Response
 processRequest req = do
   case (Httpd.reqMethod req, Network.URI.uriPath (Httpd.reqURI req)) of 
-    ("GET", "/basic/get") -> return $ Httpd.Response 200 [] "It works."
-    ("GET", "/basic/get2") -> return $ Httpd.Response 200 [] "It works (2)."
+    ("GET", "/basic/get") -> return $ Httpd.mkResponse 200 [] "It works."
+    ("GET", "/basic/get2") -> return $ Httpd.mkResponse 200 [] "It works (2)."
+    ("GET", "/basic/head") -> return $ Httpd.mkResponse 200 [] "Body for /basic/head."
+    ("HEAD", "/basic/head") -> return $ Httpd.mkResponse 200 [] "Body for /basic/head."
     ("POST", "/basic/post") ->
         let typ = lookup "Content-Type" (Httpd.reqHeaders req)
             len = lookup "Content-Length" (Httpd.reqHeaders req)
             body = Httpd.reqBody req
-        in return $ Httpd.Response 200 [] (show (typ, len, body))
+        in return $ Httpd.mkResponse 200 [] (show (typ, len, body))
 
     ("GET", "/basic/example") ->
-      return $ Httpd.Response 200 [] haskellOrgText
+      return $ Httpd.mkResponse 200 [] haskellOrgText
 
     ("GET", "/auth/basic") ->
       case lookup "Authorization" (Httpd.reqHeaders req) of
-        Just "Basic dGVzdDpwYXNzd29yZA==\r" -> return $ Httpd.Response 200 [] "Here's the secret"
-        x -> return $ Httpd.Response 401 [("WWW-Authenticate", "Basic realm=\"Testing realm\"")] (show x)
+        Just "Basic dGVzdDpwYXNzd29yZA==" -> return $ Httpd.mkResponse 200 [] "Here's the secret"
+        x -> return $ Httpd.mkResponse 401 [("WWW-Authenticate", "Basic realm=\"Testing realm\"")] (show x)
 
     ("GET", "/auth/digest") ->
       case lookup "Authorization" (Httpd.reqHeaders req) of
         Just (hasPrefix "Digest " -> Just (splitFields -> items))
-          | [("username", show "test"), ("realm", show "Digest testing realm"), ("nonce", show "87e4"),
-             ("uri", show (testUrl "/auth/digest")), ("opaque", show "057d"),
-             ("response", show "ace810a3cfb830489a3b48e90a02b2ae")] `isSubsetOf` items
-          -> return $ Httpd.Response 200 [] "Here's the digest secret"
-          | [("username", show "test"), ("realm", show "Digest testing realm"), ("nonce", show "87e4"),
-             ("uri", show "/auth/digest"), ("opaque", show "057d"),
-             ("response", show "4845c3faf4dcb125b8dcc88b5c20bb89")] `isSubsetOf` items
-          -> return $ Httpd.Response 200 [] "Here's the digest secret"
-        x -> return $ Httpd.Response
+          | digestMatch "test" "Digest testing realm" "digestpassword"
+                        "87e4" "057d"
+                        "GET" "/auth/digest" ?testUrl
+                        items
+          -> return $ Httpd.mkResponse 200 [] "Here's the digest secret"
+        x -> return $ Httpd.mkResponse
                         401
                         [("WWW-Authenticate",
                           "Digest realm=\"Digest testing realm\", opaque=\"057d\", nonce=\"87e4\"")]
                         (show x)
 
     ("GET", "/browser/example") ->
-      return $ Httpd.Response 200 [] haskellOrgText
+      return $ Httpd.mkResponse 200 [] haskellOrgText
     ("GET", "/browser/no-cookie") ->
       case lookup "Cookie" (Httpd.reqHeaders req) of
-        Nothing -> return $ Httpd.Response 200 [] ""
-        Just s  -> return $ Httpd.Response 500 [] s
+        Nothing -> return $ Httpd.mkResponse 200 [] ""
+        Just s  -> return $ Httpd.mkResponse 500 [] s
     ("GET", "/browser/one-cookie/1") ->
-      return $ Httpd.Response 200 [("Set-Cookie", "hello=world")] ""
+      return $ Httpd.mkResponse 200 [("Set-Cookie", "hello=world")] ""
     ("GET", "/browser/one-cookie/2") ->
       case lookup "Cookie" (Httpd.reqHeaders req) of
-        -- TODO: is it correct to expect the \r at the end?
-        Just "hello=world\r" -> return $ Httpd.Response 200 [] ""
-        Just s               -> return $ Httpd.Response 500 [] s
-        Nothing              -> return $ Httpd.Response 500 [] (show $ Httpd.reqHeaders req)
+        Just "hello=world" -> return $ Httpd.mkResponse 200 [] ""
+        Just s               -> return $ Httpd.mkResponse 500 [] s
+        Nothing              -> return $ Httpd.mkResponse 500 [] (show $ Httpd.reqHeaders req)
     ("GET", "/browser/two-cookies/1") ->
-      return $ Httpd.Response 200
+      return $ Httpd.mkResponse 200
                               [("Set-Cookie", "hello=world")
                               ,("Set-Cookie", "goodbye=cruelworld")]
                               ""
     ("GET", "/browser/two-cookies/2") ->
       case lookup "Cookie" (Httpd.reqHeaders req) of
-        -- TODO: is it correct to expect the \r at the end?
         -- TODO generalise the cookie parsing to allow for whitespace/ordering variations
-        Just "goodbye=cruelworld; hello=world\r" -> return $ Httpd.Response 200 [] ""
-        Just s               -> return $ Httpd.Response 500 [] s
-        Nothing              -> return $ Httpd.Response 500 [] (show $ Httpd.reqHeaders req)
+        Just "goodbye=cruelworld; hello=world" -> return $ Httpd.mkResponse 200 [] ""
+        Just s               -> return $ Httpd.mkResponse 500 [] s
+        Nothing              -> return $ Httpd.mkResponse 500 [] (show $ Httpd.reqHeaders req)
     ("GET", hasPrefix "/browser/redirect/relative/" -> Just (break (=='/') -> (maybeRead -> Just n, rest))) ->
-      return $ Httpd.Response n [("Location", rest)] ""
+      return $ Httpd.mkResponse n [("Location", rest)] ""
     ("GET", hasPrefix "/browser/redirect/absolute/" -> Just (break (=='/') -> (maybeRead -> Just n, rest))) ->
-      return $ Httpd.Response n [("Location", testUrl rest)] ""
+      return $ Httpd.mkResponse n [("Location", ?testUrl rest)] ""
     ("GET", hasPrefix "/browser/redirect/secure/" -> Just (break (=='/') -> (maybeRead -> Just n, rest))) ->
-      return $ Httpd.Response n [("Location", secureTestUrl rest)] ""
-    _                     -> return $ Httpd.Response 500 [] "Unknown request"
+      return $ Httpd.mkResponse n [("Location", ?secureTestUrl rest)] ""
+    _                     -> return $ Httpd.mkResponse 500 [] "Unknown request"
 
 altProcessRequest :: Httpd.Request -> IO Httpd.Response
 altProcessRequest req = do
   case (Httpd.reqMethod req, Network.URI.uriPath (Httpd.reqURI req)) of 
-    ("GET", "/basic/get") -> return $ Httpd.Response 200 [] "This is the alternate server."
-    ("GET", "/basic/get2") -> return $ Httpd.Response 200 [] "This is the alternate server (2)."
-    _                     -> return $ Httpd.Response 500 [] "Unknown request"
-
-getResponseCode :: Result (Response a) -> IO ResponseCode
-getResponseCode (Left err) = fail (show err)
-getResponseCode (Right r)  = return (rspCode r)
+    ("GET", "/basic/get") -> return $ Httpd.mkResponse 200 [] "This is the alternate server."
+    ("GET", "/basic/get2") -> return $ Httpd.mkResponse 200 [] "This is the alternate server (2)."
+    _                     -> return $ Httpd.mkResponse 500 [] "Unknown request"
 
 maybeTestGroup True name xs = testGroup name xs
 maybeTestGroup False name _ = testGroup name []
 
-tests port80Server =
-  [ testGroup "Basic tests"
+basicTests =
+    testGroup "Basic tests"
     [ testCase "Basic GET request" basicGetRequest
+    , testCase "Basic GET request (lazy bytestring)" basicGetRequestLBS
     , testCase "Network.HTTP example code" basicExample
     , testCase "Secure GET request" secureGetRequest
     , testCase "Basic POST request" basicPostRequest
+    , testCase "Basic HEAD request" basicHeadRequest
     , testCase "Basic Auth failure" basicAuthFailure
     , testCase "Basic Auth success" basicAuthSuccess
     ]
-  , testGroup "Browser tests"
+
+browserTests =
+    testGroup "Browser tests"
     [ testGroup "Basic"
       [
         -- github issue 14
@@ -521,7 +560,9 @@ tests port80Server =
       , testCase "Digest" browserDigestAuth
       ]
     ]
-  , maybeTestGroup port80Server "Multiple servers"
+
+port80Tests =
+    testGroup "Multiple servers"
     [ testCase "Alternate server" browserAlt
     , testCase "Both servers" browserBoth
     , testCase "Both servers (reversed)" browserBothReversed
@@ -529,13 +570,6 @@ tests port80Server =
     -- , testCase "Two requests - alternate server" browserTwoRequestsAlt
     -- , testCase "Two requests - both servers" browserTwoRequestsBoth
     ]
-  ]
-
-portNum :: Int
-portNum = 5812
-
-altPortNum :: Int
-altPortNum = 80
 
 urlRoot :: Int -> String
 urlRoot 80 = "http://localhost"
@@ -545,32 +579,49 @@ secureRoot :: Int -> String
 secureRoot 443 = "https://localhost"
 secureRoot n = "https://localhost:" ++ show n
 
-testUrl :: String -> String
-testUrl p = urlRoot portNum ++ p
+type ServerAddress = String -> String
 
-altTestUrl :: String -> String
-altTestUrl p = urlRoot altPortNum ++ p
-
-secureTestUrl :: String -> String
-secureTestUrl p = secureRoot portNum ++ p
+httpAddress, httpsAddress :: Int -> ServerAddress
+httpAddress port p = urlRoot port ++ p
+httpsAddress port p = secureRoot port ++ p
 
 main :: IO ()
 main = do
   args <- getArgs
+
+  let servers = [("httpd-shed", Httpd.shed), ("warp", Httpd.warp)]
+      basePortNum, altPortNum :: Int
+      basePortNum = 5812
+      altPortNum = 80
+      numberedServers = zip [basePortNum..] servers
+
+  let setupNormalTests = do
+      flip mapM numberedServers $ \(portNum, (serverName, server)) -> do
+         let ?testUrl = httpAddress portNum
+             ?secureTestUrl = httpsAddress portNum
+         _ <- forkIO $ server portNum processRequest
+         return $ testGroup serverName [basicTests, browserTests]
+
+  let setupAltTests = do
+      let (portNum, (_, server)) = head numberedServers
+      let ?testUrl = httpAddress portNum
+          ?altTestUrl = httpAddress altPortNum
+      _ <- forkIO $ server altPortNum altProcessRequest
+      return port80Tests
+
   case args of
      ["server"] -> do -- run only the harness servers for diagnostic/debug purposes
                       -- halt on any keypress
-        _ <- forkIO (() <$ Httpd.initServer portNum processRequest)
-        _ <- forkIO (() <$ Httpd.initServer altPortNum altProcessRequest)
+        _ <- setupNormalTests
+        _ <- setupAltTests
         _ <- getChar
         return ()
      ("--withport80":args) -> do
-        _ <- forkIO (() <$ Httpd.initServer portNum processRequest)
-        _ <- forkIO (() <$ Httpd.initServer altPortNum altProcessRequest)
+        normalTests <- setupNormalTests
+        altTests <- setupAltTests
         _ <- threadDelay 1000000 -- Give the server time to start :-(
-        defaultMainWithArgs (tests True) args
+        defaultMainWithArgs (normalTests ++ [altTests]) args
      args -> do -- run the test harness as normal
-        _ <- forkIO (() <$ Httpd.initServer portNum processRequest)
+        normalTests <- setupNormalTests
         _ <- threadDelay 1000000 -- Give the server time to start :-(
-        defaultMainWithArgs (tests False) args
-
+        defaultMainWithArgs normalTests args
